////////////////////////////////////////////////////////////////////////
// appuk.ox	replicates the empirical results reported in
// 	H. Peter Boswijk and Jurgen A. Doornik, "Distribution Approximations
// for Cointegration Tests with Stationary Exogenous Regressors".
// Journal of Applied Econometrics, forthcoming.
//
// Required: Ox 3.3 or newer
//
// uses ukppp.in7/ukppp.bn7, which is the dataset from
// Johansen and Juselius (1992):
//  sample period: 1972(1) - 1987(2)
//  frequency:     quarterly
//	p1        UK wholesale price index
//	p2        trade-weighted foreign wholesale price index
//	e12       UK effective exchange rate
//	i1        UK three-month treasury bill rate in the UK
//	i2        three-month Eurodollar interest rate in the UK
//	doilp0    change in oil price
//	Constructed variables:
//	D1980(2)  dummy for 1980(2)
//	p1-p2     p1-p2
//	Dp1       first difference of p1
//	Di2       first difference of i2
//	Cumdoilp0 cumulative sum of doilp0
////////////////////////////////////////////////////////////////////////

#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>
#import  <pcfiml>

#include <lib/coigamma.h>

////////////////////////////////////////////////////////////////////////
// Heteroscedasticity and Autocorrelation Consistent covariance matrix
// Andrews (1991), quadratic spectral kernel HAC estimator
// with plug-in automatic bandwidth using AR(1) for each component.
// adjustment: truncated when kernel weight ~< 1e-5.
HACest(const mV)
{
	decl i, ct = rows(mV), cp = columns(mV);

	// Andrews (1991), eqn (6.4) alpha2 with AR(1) for each component
	decl anumer, adenom, x, y, rho, sigma2;
	for (i = anumer = adenom = 0; i < cp; ++i)
	{
		y = mV[][i];  y[0][0] = 0;  x = lag0(mV[][i], 1);
		rho = double(y'x ./ (x'x) );
		// ct-2: one lost for lag, one for Ar parameter
		sigma2 = double(sumsqrc(y - rho * x) ) / (ct - 2);
		anumer += 4 * (rho * sigma2) ^ 2 / (1 - rho) ^ 8;
		adenom += sigma2 ^ 2 / (1 - rho) ^ 4;
	}
	// (6.2) S_T for quadratic spectral kernel
	decl st = 1.3221 * (ct * anumer / adenom) ^ 0.2;
	// determine cutoff; is at kernel argument 50, k(50)~1e-5
	decl cm = int(50 * st);
	if (cm >= ct) cm = ct - 1;
//	println("st = ", st, " alpha(2)", anumer / adenom, "cm=", cm);

	// (2.5) quadratic spectral kernel loop to compute HAC matrix
	decl hac, gamma, pi65 = 6 * M_PI / 5, pisqr = 25 / (12 * M_PI^2), pi65x, kern;
	for (i = -cm, hac = 0; i <= cm; ++i)
	{
		x = i / st;
		if (fabs(x) <= 0.0001)
			kern = 1;
		else
		{	pi65x = pi65 * x;
			kern = (sin(pi65x) / pi65x - cos(pi65x)) * pisqr / sqr(x);
		}
		gamma = i < 0  ?  lag0(mV, -i)'mV  :  mV'lag0(mV, i);
		hac += kern * gamma;
//		println("i=", "%3d", i, " x=", x, " kern=", kern, " sum=", st);
	}
	return hac / ct;
}
////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
// extension: stationary regressors in cointegration analysis
class PcFimlExt : PcFiml
{
	PcFimlExt();
	CointegrationI0(const cI0, const bUseHac);
	CointegrationI1Cond(const cExtra);
};

PcFimlExt::PcFimlExt()
{
	PcFiml();
}
PcFimlExt::CointegrationI0(const cI0, const bUseHac)
{
    decl xu = GetGroup(U_VAR)[][ : cI0 - 1];
	decl bu = m_mPi[][m_cW - m_cU : m_cW - m_cU + cI0 - 1];
	decl u = m_mResidual;
	decl v = m_mResidual + xu * bu';
	decl i, j, cp, aorth, ua, va, eval, evec, hac, icase, gammapar,
		pval, rho, pvalot, pvalt;

	icase = GetStatus(&i, &j);
	gammapar = GammaParTraceAsymp(icase, range(m_cY,1)', m_cT - m_cW);
	pvalot = 1 - probgamma(m_mEval[1][]', gammapar[][0], gammapar[][1]);

	println("\nCointegration tests with I(0) variables", bUseHac ? "(HAC)" : "");
	println("rank", "%12s", "test", "    pval ", " pval(T) ", "pvalorig", sprint("  orig T=", m_cT - m_cW));

	for (i = 0, rho = <>; i < m_cY; ++i)
	{
		aorth = i == 0 ? unit(m_cY) : nullspace(m_mAlpha[][ : i - 1]);
		//if (i > 0) print(aorth, m_mAlpha[][:i-1]'aorth);
		ua = u * aorth;
		va = v * aorth;
		cp = columns(aorth);

		if (bUseHac)
		{	hac = HACest(ua ~ va);
			eigensymgen(hac[:cp-1][cp:] * invertsym(hac[cp:][cp:]) * hac[cp:][:cp-1],
				hac[:cp-1][:cp-1], &eval, &evec);
		}
		else
			eigensymgen(ua'va * invertsym(va'va) * va'ua, ua'ua, &eval, &evec);

		eval = sqrt(eval);
		gammapar = GammaParI0CondAsymp(icase, eval, 0);
		pval = 1 - probgamma(m_mEval[1][i], gammapar[][0], gammapar[][1]);
		gammapar = GammaParI0CondAsymp(icase, eval, m_cT - m_cW);
		pvalt = 1 - probgamma(m_mEval[1][i], gammapar[][0], gammapar[][1]);
		
		println("%4d", i, "%#12.6g", m_mEval[1][i], " [", "%6.4f", pval, "]",
			" [", "%6.4f", pvalt, "]",
			" [", "%6.4f", m_mEval[2][i], "]", " [", "%6.4f", pvalot[i], "]");
		eval ~= constant(M_NAN, 1, i);
		rho |= eval;
	}
	print("rho", rho);
}
PcFimlExt::CointegrationI1Cond(const cExtra)
{
	decl i, j, icase, gammapar, pval, pvalt;

	icase = GetStatus(&i, &j);
	gammapar = GammaParI1CondAsymp(icase, cExtra + range(m_cY,1)', range(m_cY,1)', 0);
	pval = 1 - probgamma(m_mEval[1][]', gammapar[][0], gammapar[][1]);
	gammapar = GammaParI1CondAsymp(icase, cExtra + range(m_cY,1)', range(m_cY,1)', m_cT - m_cW);
	pvalt = 1 - probgamma(m_mEval[1][]', gammapar[][0], gammapar[][1]);

	println("\nCointegration tests conditional on ", cExtra, " I(1) variable(s)");
	println("rank", "%14s", "test", "    pval ", " pval(T) ", "pvalorig");

	for (i = 0; i < m_cY; ++i)
		println("%4d", i, "%#14.6g", m_mEval[1][i], " [", "%6.4f", pval[i], "]",
			" [", "%6.4f", pvalt[i], "]"," [", "%6.4f", m_mEval[2][i], "]");
}
////////////////////////////////////////////////////////////////////////


main()
{
    decl system;

    system = new PcFimlExt();

    system->Load("ukppp.in7");
    system->Deterministic(TRUE);
                                    // formulate the system
    system->Select(Y_VAR, { "p1-p2",0,2, "e12",0,2, "Dp1",0,2, "i1",0,2, "i2",0,2 } );
    system->Select(X_VAR, { "Trend", 0, 0 } );
    system->Select(U_VAR, { "doilp0", 0, 1 } );
    system->Select(U_VAR, { "Constant", 0, 0 } );
    system->Select(U_VAR, { "CSeason", 0, 2 } );

    system->SetSelSample(-1, 1, -1, 1);
    system->Estimate();        // estimate the system (VAR)
    system->Cointegration();   // cointegration analysis
	println("\n=== Table 1");
    system->CointegrationI0(2, TRUE);	// first two U_VARs are additional I(0) variables
	println("\n=== non-HAC is nearly identical:");
    system->CointegrationI0(2, FALSE);	// first two U_VARs are additional I(0) variables
	system->CointegrationI1Cond(1);

    system->Select(X_VAR, { "Cumdoilp0", 1, 1 } );
    system->SetSelSample(-1, 1, -1, 1);
    system->Estimate();        // estimate the system (VAR)
    system->Cointegration();   // cointegration analysis
	println("\n=== Table 2");
	system->CointegrationI1Cond(1);

	system->DeSelect();
    system->Select(Y_VAR, { "p1-p2",0,1, "e12",0,1, "Dp1",0,1, "i1",0,1, "i2",0,1 } );
    system->Select(X_VAR, { "Trend", 0, 0 } );
    system->Select(U_VAR, { "doilp0", 0, 1 } );
    system->Select(U_VAR, { "Constant", 0, 0 } );
    system->Select(U_VAR, { "Di2", 1, 1 } );
    system->Select(U_VAR, { "D1980(2)", 0, 0 } );

    system->SetSelSample(1972, 4, -1, 1);
    system->Estimate();        // estimate the system (VAR)
    system->Cointegration();   // cointegration analysis
	println("\n=== Table 3 (columns 3,4,5)");
    system->CointegrationI0(2, TRUE);	// first two U_VARs are additional I(0) variables
    system->CointegrationI0(2, FALSE);	// first two U_VARs are additional I(0) variables

    system->Select(X_VAR, { "Cumdoilp0", 1, 1 } );
    system->SetSelSample(1972, 4, -1, 1);
    system->Estimate();        // estimate the system (VAR)
    system->Cointegration();   // cointegration analysis
	println("\n=== Table 3 (columns 6,7)");
	system->CointegrationI1Cond(1);
	
	delete system;
}

